#version 400 compatibility

/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/*DRAWBUFFERS:023*/
layout(location = 0) out vec4 scene;
layout(location = 1) out vec4 gData1;
layout(location = 2) out vec4 out3;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/colorspace.glsl"

const int shadowMapResolution   = 2560;     //[512 1024 1536 2048 2560 3072 3584 4096 6144 8192 16384]
const float shadowDistance      = 128.0;

const bool shadowHardwareFiltering = false;

in vec2 coord;

flat in mat4x3 lightColor;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;

uniform sampler2D noisetex;

uniform sampler2D depthtex0;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;

uniform int frameCounter;
uniform int isEyeInWater;

uniform float aspectRatio;
uniform float far, near;
uniform float lightFlip;
uniform float sunAngle;
uniform float viewHeight, viewWidth;

uniform vec2 taaOffset;
uniform vec2 viewSize, pixelSize;

uniform vec3 lightvec, lightvecView;
uniform vec3 upvec;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;


/* ------ includes ------ */
#include "/lib/util/transforms.glsl"

#define FUTIL_LINDEPTH
#define FUTIL_MAT16
#define FUTIL_LIGHTMAP
#include "/lib/fUtil.glsl"

#include "/lib/light/diffuse.glsl"

#include "/lib/light/warp.glsl"

#include "/lib/frag/bluenoise.glsl"
#include "/lib/frag/gradnoise.glsl"

#define SOLIDPASS

#include "/lib/light/shadow.glsl"

#include "/lib/atmos/phase.glsl"

#include "/lib/light/subsurface.glsl"

#include "/lib/light/contactShadow.glsl"

#include "/lib/light/brdf.glsl"

#include "/lib/frag/labPBR.glsl"


/* ------ metal albedo encoding ------ */

float bayer2e(vec2 a){
    a = floor(a);
    return fract( dot(a, vec2(.5, a.y * .75)) );
}
#define bayer4e(a)   (bayer2e( .5*(a))*.25+bayer2e(a))

#define m vec3(31,63,31)
float encode3x8(vec3 a){
    float dither = bayer4e(gl_FragCoord.xy);
    a += (dither-.5) / m;
    a = saturate(a);
    ivec3 b = ivec3(a*m);
    return float( b.r|(b.g<<5)|(b.b<<11) ) / 65535.;
}
#undef m

vec4 textureBilateral(sampler2D tex, sampler2D depth, const int lod, float fdepth) {
    vec4 data   = vec4(0.0);
    float sum   = 0.0;
    ivec2 posD  = ivec2(coord*viewSize);
    ivec2 posT  = ivec2(coord*viewSize*rcp(float(lod)));
    vec3 zmult  = vec3((far*near)*2.0, far+near, far-near);
        fdepth  = depthLinear(fdepth);
    
    for (int i = -1; i<2; i++) {
        for (int j = -1; j<2; j++) {
            ivec2 tcDepth = posD + ivec2(i, j)*lod;
            float dsample = depthLinear(texelFetch(depth, tcDepth, 0).x);
            float w     = abs(dsample-fdepth)*zmult.x<1.0 ? 1.0 : 1e-5;
            ivec2 ct    = posT + ivec2(i, j);
            data       += texelFetch(tex, ct, 0)*w;
            sum        += w;
        }
    }
    data *= rcp(sum);

    return data;
}

vec3 fauxPorosity(vec3 albedo, float wetness, float porosity) {
    vec3 wetAlbedo      = colorSaturation(mix(albedo * sqrt(albedo), sqr(albedo), getLuma(albedo)), 0.85) * 0.95;

    return mix(albedo, wetAlbedo, wetness * porosity);
}

void main() {
    vec4 sceneColor = stex(colortex0);
    float sceneDepth = stex(depthtex0).x;
    vec4 tex2       = stex(colortex2);

    float wetness   = 0.0;

    vec3 albedo     = sceneColor.rgb;

    if (landMask(sceneDepth)) {
        vec4 tex1       = stex(colortex1);
        vec2 sceneLmap  = decode2x8(tex1.z);
        vec3 sceneNormal = decodeNormal(tex1.xy);
        vec3 viewNormal = mat3(gbufferModelView) * sceneNormal;

        int matID       = decodeMatID16(tex2.x);

        vec4 specularData = vec4(decode2x8(tex2.y), decode2x8(tex2.z));
        vec2 aux        = decode2x8(tex2.w);

        materialLAB matLAB = decodeSpecularTexture(specularData);

        #ifndef resourcepackReflectionsEnabled
            matLAB.porosity = 0.6;
        #endif
            wetness     = tex1.w;

        sceneColor.rgb  = fauxPorosity(sceneColor.rgb, wetness, matLAB.porosity);

        #if DEBUG_VIEW==1
        sceneColor.rgb    = vec3(0.25);
        #endif

        vec3 viewPos    = screenToViewSpace(vec3(coord, sceneDepth));
        vec3 viewDir    = normalize(viewPos);
        vec3 scenePos   = viewToSceneSpace(viewPos);

        #ifdef lightmapSmoothingEnabled
        vec3 tex3       = stex(colortex3).xyz;
            sceneLmap   = tex3.xy;
            sceneColor.a = tex3.z;
        #endif

        float ao        = sqr(sceneColor.a)*0.94+0.06;
        
        #ifdef ambientOcclusionEnabled
            vec4 upscale3 = textureBilateral(colortex3, depthtex0, 2, sceneDepth);
                ao     *= upscale3.w;
        #endif
        
        /* ------ lighting calculation ------ */
        //~6 fps
        
        vec2 lmap       = sceneLmap;
        lmap.y          = cube(lmap.y);

        #if diffuseModel == 0
        float diff      = diffuseHammon(viewNormal, -viewDir, lightvecView, matLAB.roughness);
        #else
        float diff      = diffuseLambert(sceneNormal, lightvec);
        #endif

        #ifndef subsurfaceScatteringEnabled
            if (matID == 2) diff = diff*0.3+0.7 / pi;
            if (matID == 4) diff = sqrt(diff);
        #endif

        shadowData shadow = shadowData(1.0, vec3(1.0), vec3(0.0));

        #ifdef subsurfaceScatteringEnabled
            if (matID == 2 || matID == 4) shadow = getShadowSubsurface(diff>0.0, scenePos, normalize(viewPos), sceneColor.rgb, 1.0);
            else shadow = getShadowRegular(diff>0.0, scenePos);    //~3fps, almost for free
        #else
            shadow = getShadowRegular(diff>0.0, scenePos);    //~3fps, almost for free
        #endif

        float diffLit       = diff * shadow.shadow * aux.y;

        #ifdef contactShadowsEnabled
            if (diffLit > 0.0 && isEyeInWater == 0) diffLit *= getContactShadow(depthtex0, viewPos, ditherBluenoise(), sceneDepth, dot(viewNormal, viewDir), lightvecView);
        #endif

        vec3 directCol      = (sunAngle<0.5 ? lightColor[0] : lightColor[2]) * lightFlip;
        vec3 directLight    = diffLit * shadow.color * directCol;
        vec3 specDirectLight = directLight;

        #ifdef subsurfaceScatteringEnabled
            directLight    += shadow.subsurfaceScatter * directCol;
        #endif

            directLight    *= ao * 0.9 + 0.1;

        vec3 indirectLight  = lmap.y*lightColor[1];
            indirectLight  += vec3(0.5, 0.7, 1.0) * (0.07 * minLightLuma);
            indirectLight  += directCol * cube(lmap.y) * 0.03 * saturate(dot(sceneNormal, -upvec) * 0.5 + 0.5);
            indirectLight  *= ao;

        vec3 result     = directLight + indirectLight * aux.x;
            result     += getBlocklightMap(lightColor[3], lmap.x)*ao;
        
        #ifdef labEmissionEnabled
            result     += matLAB.emission * v3avg(lightColor[3]);
        #endif

        if (matID == 5) {
            vec3 emitterCol     = mix(lightColor[3], vec3(v3avg(lightColor[3])), emitterRecoloring);
                emitterCol     *= cubeSmooth(saturate(v3avg(sceneColor.rgb) * 1.5));
            result += emitterCol * pi;
        }

        sceneColor.rgb   *= result;
        //sceneColor.rgb    = shadow.subsurfaceScatter;
        //sceneColor.rgb  = vec3(getContactShadow(depthtex0, viewPos, ditherBluenoise(), sceneDepth, dot(viewNormal, viewDir)));

        #ifdef specularBRDFEnabled
        if (diffLit > 0.0) {
            vec3 brdf   = BRDF(-viewDir, lightvecView, viewNormal, labToGeneric(matLAB), albedo) * specDirectLight;

            #ifdef lightleakWorkaroundToggle
                brdf   *= sstep(sceneLmap.y, 0.1, 0.2);
            #endif

            //vec3 fresnel = BRDFfresnel(-viewDir, viewNormal, material);
            sceneColor.rgb = sceneColor.rgb + brdf;
            //sceneColor.rgb *= fresnel;
        }
        #endif

        #if DEBUG_VIEW==2
        sceneColor.rgb    = ao * (sunAngle<0.5 ? lightColor[0] : lightColor[3]);
        #endif
    }

    tex2.z  = encode3x8(sqrt(saturate(albedo.rgb)));

    scene  = makeDrawbuffer(sceneColor);
    gData1 = clamp16F(tex2);
    out3  = vec4(0.0);
}